#!/usr/bin/env python

# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>
#
# This script automatically generates the bodies of the following files, creating the full CLBlast API interface and
# implementation (C, C++, and reference BLAS wrappers):
#    clblast.h
#    clblast.cpp
#    clblast_c.h
#    clblast_c.cpp
#    clblast_cuda.h
#    clblast_cuda.cpp
#    clblast_netlib_c.h
#    clblast_netlib_c.cpp
#    wrapper_clblas.h
#    wrapper_cblas.h
#    pyclblast.pyx
# It also generates the main functions for the correctness and performance tests as found in
#    test/correctness/routines/levelX/xYYYY.cpp
#    test/performance/routines/levelX/xYYYY.cpp
# It also produces the API documentation found in doc/clblast.md


import sys
import os.path
import argparse

import generator.cpp as cpp
import generator.doc as doc
import generator.pyclblast as pyclblast
from generator.routine import Routine
from generator.datatype import H, S, D, C, Z, Sc, Dz, iH, iS, iD, iC, iZ, Css, Zdd, Ccs, Zzd, T, Tc, TU

FILES = [
    "/include/clblast.h",
    "/src/clblast.cpp",
    "/include/clblast_c.h",
    "/src/clblast_c.cpp",
    "/test/wrapper_clblas.hpp",
    "/test/wrapper_cblas.hpp",
    "/test/wrapper_cublas.hpp",
    "/include/clblast_netlib_c.h",
    "/src/clblast_netlib_c.cpp",
    "/include/clblast_cuda.h",
    "/src/clblast_cuda.cpp",
    "/src/pyclblast/src/pyclblast.pyx"
]
HEADER_LINES = [129, 21, 133, 24, 29, 45, 29, 66, 40, 96, 21, 327]
FOOTER_LINES = [98, 57, 112, 275, 6, 6, 6, 9, 2, 41, 56, 37]
HEADER_LINES_DOC = 0
FOOTER_LINES_DOC = 232

# Different possibilities for requirements
ald_m = "The value of `a_ld` must be at least `m`."
ald_n = "The value of `a_ld` must be at least `n`."
ald_k_one = "The value of `a_ld` must be at least `k + 1`."
ald_kl_ku_one = "The value of `a_ld` must be at least `kl + ku + 1`."
ald_transa_m_k = "When `(transpose_a == Transpose::kNo && layout == Layout::kColMajor) || (transpose_a == Transpose::kYes && layout == Layout::kRowMajor)`, then `a_ld` must be at least `m`, otherwise `a_ld` must be at least `k`."
ald_trans_n_k = "When `(transpose == Transpose::kNo && layout == Layout::kColMajor) || (transpose == Transpose::kYes && layout == Layout::kRowMajor)`, then `a_ld` must be at least `n`, otherwise `a_ld` must be at least `k`."
ald_side_m_n = "When `side = Side::kLeft` then `a_ld` must be at least `m`, otherwise `a_ld` must be at least `n`."
bld_m = "The value of `b_ld` must be at least `m`."
bld_n = "The value of `b_ld` must be at least `n`."
bld_transb_k_n = "When `(transpose_b == Transpose::kNo && layout == Layout::kColMajor) || (transpose_b == Transpose::kYes && layout == Layout::kRowMajor)`, then `b_ld` must be at least `k`, otherwise `b_ld` must be at least `n`."
bld_trans_n_k = "When `(transpose == Transpose::kNo && layout == Layout::kColMajor) || (transpose == Transpose::kYes && layout == Layout::kRowMajor)`, then `b_ld` must be at least `n`, otherwise `b_ld` must be at least `k`."
cld_m = "The value of `c_ld` must be at least `m`."
cld_n = "The value of `c_ld` must be at least `n`."


# Helper functions to compute vector and matrix sizes
def size_helper(condition, size_one, size_two, multiplier):
    length = "(" + condition + ")" + " ? " + size_one + " * " + multiplier + " : " + size_two + " * " + multiplier
    return length


def layout_transpose_condition(prefix):
    return "(layout == CLBlastLayoutColMajor && " + prefix + "_transpose != CLBlastTransposeNo) || " +\
           "(layout == CLBlastLayoutRowMajor && " + prefix + "_transpose == CLBlastTransposeNo)"


# Different possibilities for the vector and matrix sizes
xn = "n * x_inc"
xm = "m * x_inc"
yn = "n * y_inc"
ym = "m * y_inc"
zn = "n * z_inc"
an = "n * a_ld"
apn = "((n*(n+1)) / 2)"
cn = "n * c_ld"
xmn = size_helper("a_transpose != CLBlastTransposeNo", "m", "n", "x_inc")
ynm = size_helper("a_transpose != CLBlastTransposeNo", "n", "m", "y_inc")
amn = size_helper("layout == CLBlastLayoutRowMajor", "m", "n", "a_ld")
amns = size_helper("side == CLBlastSideLeft", "m", "n", "a_ld")
amk = size_helper(layout_transpose_condition("a"), "m", "k", "a_ld")
ank = size_helper(layout_transpose_condition("a"), "n", "k", "a_ld")
ankab = size_helper(layout_transpose_condition("ab"), "n", "k", "a_ld")
bkn = size_helper(layout_transpose_condition("b"), "k", "n", "b_ld")
bnkab = size_helper(layout_transpose_condition("ab"), "n", "k", "b_ld")
bmn = size_helper("layout == CLBlastLayoutRowMajor", "m", "n", "b_ld")
bnma = size_helper(layout_transpose_condition("a"), "n", "m", "b_ld")
cmn = size_helper("layout == CLBlastLayoutRowMajor", "m", "n", "c_ld")
ammn = size_helper("layout == CLBlastLayoutRowMajor", "m", "((side == CLBlastSideLeft) ? m : n)", "a_ld")
bmnn = size_helper("layout == CLBlastLayoutRowMajor", "((side == CLBlastSideLeft) ? m : n)", "n", "b_ld")
im = "height * width * channels"
col = "height * width * channels"
imb = "height * width * channels * batch_count"
kernel = "kernel_h * kernel_w * num_kernels"
result = "height_out * width_out * num_kernels * batch_count"


# ==================================================================================================

# Populates a list of routines
im2col_constants = ["channels", "height", "width", "kernel_h", "kernel_w", "pad_h", "pad_w", "stride_h", "stride_w", "dilation_h", "dilation_w"]
convgemm_constants = im2col_constants + ["num_kernels", "batch_count"]
ROUTINES = [
[  # Level 1: vector-vector
  Routine(False, True,  0, False, "1", "rotg",  T, [S,D],            [],                  [],                                                     [],         ["sa","sb","sc","ss"],        ["1","1","1","1"], [],       "",    "Generate givens plane rotation", "", []),
  Routine(False, True,  0, False, "1", "rotmg", T, [S,D],            [],                  [],                                                     ["sy1"],    ["sd1","sd2","sx1","sparam"], ["1","1","1","1","1"], [],   "",    "Generate modified givens plane rotation", "", []),
  Routine(False, True,  0, False, "1", "rot",   T, [S,D],            ["n"],               [],                                                     [],         ["x","y"],                    [xn,yn],       ["cos","sin"],"",    "Apply givens plane rotation", "", []),
  Routine(False, True,  0, False, "1", "rotm",  T, [S,D],            ["n"],               [],                                                     [],         ["x","y","sparam"],           [xn,yn,"1"],   [],           "",    "Apply modified givens plane rotation", "", []),
  Routine(True,  True,  0, False, "1", "swap",  T, [S,D,C,Z,H],      ["n"],               [],                                                     [],         ["x","y"],                    [xn,yn],       [],           "",    "Swap two vectors", "Interchanges _n_ elements of vectors _x_ and _y_.", []),
  Routine(True,  True,  0, False, "1", "scal",  T, [S,D,C,Z,H],      ["n"],               [],                                                     [],         ["x"],                        [xn],          ["alpha"],    "",    "Vector scaling", "Multiplies _n_ elements of vector _x_ by a scalar constant _alpha_.", []),
  Routine(True,  True,  0, False, "1", "copy",  T, [S,D,C,Z,H],      ["n"],               [],                                                     ["x"],      ["y"],                        [xn,yn],       [],           "",    "Vector copy", "Copies the contents of vector _x_ into vector _y_.", []),
  Routine(True,  True,  0, False, "1", "axpy",  T, [S,D,C,Z,H],      ["n"],               [],                                                     ["x"],      ["y"],                        [xn,yn],       ["alpha"],    "",    "Vector-times-constant plus vector", "Performs the operation _y = alpha * x + y_, in which _x_ and _y_ are vectors and _alpha_ is a scalar constant.", []),
  Routine(True,  True,  0, False, "1", "dot",   T, [S,D,H],          ["n"],               [],                                                     ["x","y"],  ["dot"],                      [xn,yn,"1"],   [],           "n",   "Dot product of two vectors", "Multiplies _n_ elements of the vectors _x_ and _y_ element-wise and accumulates the results. The sum is stored in the _dot_ buffer.", []),
  Routine(True,  True,  0, False, "1", "dotu",  T, [C,Z],            ["n"],               [],                                                     ["x","y"],  ["dot"],                      [xn,yn,"1"],   [],           "n",   "Dot product of two complex vectors", "See the regular xDOT routine.", []),
  Routine(True,  True,  0, False, "1", "dotc",  T, [C,Z],            ["n"],               [],                                                     ["x","y"],  ["dot"],                      [xn,yn,"1"],   [],           "n",   "Dot product of two complex vectors, one conjugated", "See the regular xDOT routine.", []),
  Routine(True,  True,  0, False, "1", "nrm2",  T, [S,D,Sc,Dz,H],    ["n"],               [],                                                     ["x"],      ["nrm2"],                     [xn,"1"],      [],           "2*n", "Euclidian norm of a vector", "Accumulates the square of _n_ elements in the _x_ vector and takes the square root. The resulting L2 norm is stored in the _nrm2_ buffer.", []),
  Routine(True,  True,  0, False, "1", "asum",  T, [S,D,Sc,Dz,H],    ["n"],               [],                                                     ["x"],      ["asum"],                     [xn,"1"],      [],           "n",   "Absolute sum of values in a vector", "Accumulates the absolute value of _n_ elements in the _x_ vector. The results are stored in the _asum_ buffer.", []),
  Routine(True,  False, 0, False, "1", "sum",   T, [S,D,Sc,Dz,H],    ["n"],               [],                                                     ["x"],      ["sum"],                      [xn,"1"],      [],           "n",   "Sum of values in a vector (non-BLAS function)", "Accumulates the values of _n_ elements in the _x_ vector. The results are stored in the _sum_ buffer. This routine is the non-absolute version of the xASUM BLAS routine.", []),
  Routine(True,  True,  0, False, "1", "amax",  T, [iS,iD,iC,iZ,iH], ["n"],               [],                                                     ["x"],      ["imax"],                     [xn,"1"],      [],           "2*n", "Index of absolute maximum value in a vector", "Finds the index of a maximum (not necessarily the first if there are multiple) of the absolute values in the _x_ vector. The resulting integer index is stored in the _imax_ buffer.", []),
  Routine(True,  False, 0, False, "1", "amin",  T, [iS,iD,iC,iZ,iH], ["n"],               [],                                                     ["x"],      ["imin"],                     [xn,"1"],      [],           "2*n", "Index of absolute minimum value in a vector (non-BLAS function)", "Finds the index of a minimum (not necessarily the first if there are multiple) of the absolute values in the _x_ vector. The resulting integer index is stored in the _imin_ buffer.", []),
  Routine(True,  False, 0, False, "1", "max",   T, [iS,iD,iC,iZ,iH], ["n"],               [],                                                     ["x"],      ["imax"],                     [xn,"1"],      [],           "2*n", "Index of maximum value in a vector (non-BLAS function)", "Finds the index of a maximum (not necessarily the first if there are multiple) of the values in the _x_ vector. The resulting integer index is stored in the _imax_ buffer. This routine is the non-absolute version of the IxAMAX BLAS routine.", []),
  Routine(True,  False, 0, False, "1", "min",   T, [iS,iD,iC,iZ,iH], ["n"],               [],                                                     ["x"],      ["imin"],                     [xn,"1"],      [],           "2*n", "Index of minimum value in a vector (non-BLAS function)", "Finds the index of a minimum (not necessarily the first if there are multiple) of the values in the _x_ vector. The resulting integer index is stored in the _imin_ buffer. This routine is the non-absolute minimum version of the IxAMAX BLAS routine.", []),
],
[  # Level 2: matrix-vector
  Routine(True,  True,  0, False, "2a", "gemv",  T,  [S,D,C,Z,H],    ["m","n"],           ["layout","a_transpose"],                               ["a","x"],  ["y"],                        [amn,xmn,ynm], ["alpha","beta"], "",    "General matrix-vector multiplication", "Performs the operation _y = alpha * A * x + beta * y_, in which _x_ is an input vector, _y_ is an input and output vector, _A_ is an input matrix, and _alpha_ and _beta_ are scalars. The matrix _A_ can optionally be transposed before performing the operation.", [ald_m]),
  Routine(True,  True,  0, False, "2a", "gbmv",  T,  [S,D,C,Z,H],    ["m","n","kl","ku"], ["layout","a_transpose"],                               ["a","x"],  ["y"],                        [amn,xmn,ynm], ["alpha","beta"], "",    "General banded matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is banded instead.", [ald_kl_ku_one]),
  Routine(True,  True,  0, False, "2a", "hemv",  T,  [C,Z],          ["n"],               ["layout","triangle"],                                  ["a","x"],  ["y"],                        [an,xn,yn],    ["alpha","beta"], "",    "Hermitian matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is an Hermitian matrix instead.", [ald_n]),
  Routine(True,  True,  0, False, "2a", "hbmv",  T,  [C,Z],          ["n","k"],           ["layout","triangle"],                                  ["a","x"],  ["y"],                        [an,xn,yn],    ["alpha","beta"], "",    "Hermitian banded matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is an Hermitian banded matrix instead.", [ald_k_one]),
  Routine(True,  True,  0, False, "2a", "hpmv",  T,  [C,Z],          ["n"],               ["layout","triangle"],                                  ["ap","x"], ["y"],                        [apn,xn,yn],   ["alpha","beta"], "",    "Hermitian packed matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is an Hermitian packed matrix instead and represented as _AP_.", []),
  Routine(True,  True,  0, False, "2a", "symv",  T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["a","x"],  ["y"],                        [an,xn,yn],    ["alpha","beta"], "",    "Symmetric matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is symmetric instead.", [ald_n]),
  Routine(True,  True,  0, False, "2a", "sbmv",  T,  [S,D,H],        ["n","k"],           ["layout","triangle"],                                  ["a","x"],  ["y"],                        [an,xn,yn],    ["alpha","beta"], "",    "Symmetric banded matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is symmetric and banded instead.", [ald_k_one]),
  Routine(True,  True,  0, False, "2a", "spmv",  T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["ap","x"], ["y"],                        [apn,xn,yn],   ["alpha","beta"], "",    "Symmetric packed matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is a symmetric packed matrix instead and represented as _AP_.", []),
  Routine(True,  True,  0, False, "2a", "trmv",  T,  [S,D,C,Z,H],    ["n"],               ["layout","triangle","a_transpose","diagonal"],         ["a"],      ["x"],                        [an,xn],       [],               "n",   "Triangular matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is triangular instead.", [ald_n]),
  Routine(True,  True,  0, False, "2a", "tbmv",  T,  [S,D,C,Z,H],    ["n","k"],           ["layout","triangle","a_transpose","diagonal"],         ["a"],      ["x"],                        [an,xn],       [],               "n",   "Triangular banded matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is triangular and banded instead.", [ald_k_one]),
  Routine(True,  True,  0, False, "2a", "tpmv",  T,  [S,D,C,Z,H],    ["n"],               ["layout","triangle","a_transpose","diagonal"],         ["ap"],     ["x"],                        [apn,xn],      [],               "n",   "Triangular packed matrix-vector multiplication", "Same operation as xGEMV, but matrix _A_ is a triangular packed matrix instead and repreented as _AP_.", []),
  Routine(True,  True,  0, False, "2a", "trsv",  T,  [S,D,C,Z],      ["n"],               ["layout","triangle","a_transpose","diagonal"],         ["a"],      ["x"],                        [an,xn],       [],               "",    "Solves a triangular system of equations", "", []),
  Routine(False, True,  0, False, "2a", "tbsv",  T,  [S,D,C,Z],      ["n","k"],           ["layout","triangle","a_transpose","diagonal"],         ["a"],      ["x"],                        [an,xn],       [],               "",    "Solves a banded triangular system of equations", "", [ald_k_one]),
  Routine(False, True,  0, False, "2a", "tpsv",  T,  [S,D,C,Z],      ["n"],               ["layout","triangle","a_transpose","diagonal"],         ["ap"],     ["x"],                        [apn,xn],      [],               "",    "Solves a packed triangular system of equations", "", []),
  # Level 2: matrix update
  Routine(True,  True,  0, False, "2b", "ger",   T,  [S,D,H],        ["m","n"],           ["layout"],                                             ["x","y"],  ["a"],                        [xm,yn,amn],   ["alpha"],        "",    "General rank-1 matrix update", "Performs the operation _A = alpha * x * y^T + A_, in which _x_ is an input vector, _y^T_ is the transpose of the input vector _y_, _A_ is the matrix to be updated, and _alpha_ is a scalar value.", [ald_m]),
  Routine(True,  True,  0, False, "2b", "geru",  T,  [C,Z],          ["m","n"],           ["layout"],                                             ["x","y"],  ["a"],                        [xm,yn,amn],   ["alpha"],        "",    "General rank-1 complex matrix update", "Same operation as xGER, but with complex data-types.", [ald_m]),
  Routine(True,  True,  0, False, "2b", "gerc",  T,  [C,Z],          ["m","n"],           ["layout"],                                             ["x","y"],  ["a"],                        [xm,yn,amn],   ["alpha"],        "",    "General rank-1 complex conjugated matrix update", "Same operation as xGERU, but the update is done based on the complex conjugate of the input vectors.", [ald_m]),
  Routine(True,  True,  0, False, "2b", "her",   Tc, [Css,Zdd],      ["n"],               ["layout","triangle"],                                  ["x"],      ["a"],                        [xn,an],       ["alpha"],        "",    "Hermitian rank-1 matrix update", "Performs the operation _A = alpha * x * x^T + A_, in which x is an input vector, x^T is the transpose of this vector, _A_ is the triangular Hermetian matrix to be updated, and alpha is a scalar value.", [ald_n]),
  Routine(True,  True,  0, False, "2b", "hpr",   Tc, [Css,Zdd],      ["n"],               ["layout","triangle"],                                  ["x"],      ["ap"],                       [xn,apn],      ["alpha"],        "",    "Hermitian packed rank-1 matrix update", "Same operation as xHER, but matrix _A_ is an Hermitian packed matrix instead and represented as _AP_.", []),
  Routine(True,  True,  0, False, "2b", "her2",  T,  [C,Z],          ["n"],               ["layout","triangle"],                                  ["x","y"],  ["a"],                        [xn,yn,an],    ["alpha"],        "",    "Hermitian rank-2 matrix update", "Performs the operation _A = alpha * x * y^T + conj(alpha) * y * x^T + A_, in which _x_ is an input vector and _x^T_ its transpose, _y_ is an input vector and _y^T_ its transpose, _A_ is the triangular Hermetian matrix to be updated, _alpha_ is a scalar value and _conj(alpha)_ its complex conjugate.", [ald_n]),
  Routine(True,  True,  0, False, "2b", "hpr2",  T,  [C,Z],          ["n"],               ["layout","triangle"],                                  ["x","y"],  ["ap"],                       [xn,yn,apn],   ["alpha"],        "",    "Hermitian packed rank-2 matrix update", "Same operation as xHER2, but matrix _A_ is an Hermitian packed matrix instead and represented as _AP_.", []),
  Routine(True,  True,  0, False, "2b", "syr",   T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["x"],      ["a"],                        [xn,an],       ["alpha"],        "",    "Symmetric rank-1 matrix update", "Same operation as xHER, but matrix A is a symmetric matrix instead.", [ald_n]),
  Routine(True,  True,  0, False, "2b", "spr",   T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["x"],      ["ap"],                       [xn,apn],      ["alpha"],        "",    "Symmetric packed rank-1 matrix update", "Same operation as xSPR, but matrix _A_ is a symmetric packed matrix instead and represented as _AP_.", []),
  Routine(True,  True,  0, False, "2b", "syr2",  T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["x","y"],  ["a"],                        [xn,yn,an],    ["alpha"],        "",    "Symmetric rank-2 matrix update", "Same operation as xHER2, but matrix _A_ is a symmetric matrix instead.", [ald_n]),
  Routine(True,  True,  0, False, "2b", "spr2",  T,  [S,D,H],        ["n"],               ["layout","triangle"],                                  ["x","y"],  ["ap"],                       [xn,yn,apn],   ["alpha"],        "",    "Symmetric packed rank-2 matrix update", "Same operation as xSPR2, but matrix _A_ is a symmetric packed matrix instead and represented as _AP_.", []),
],
[  # Level 3: matrix-matrix
  Routine(True,  True,  0, True,  "3", "gemm",  T,  [S,D,C,Z,H],     ["m","n","k"],        ["layout","a_transpose","b_transpose"],                ["a","b"],  ["c"],                        [amk,bkn,cmn],   ["alpha","beta"], "",    "General matrix-matrix multiplication", "Performs the matrix product _C = alpha * A * B + beta * C_, in which _A_ (_m_ by _k_) and _B_ (_k_ by _n_) are two general rectangular input matrices, _C_ (_m_ by _n_) is the matrix to be updated, and _alpha_ and _beta_ are scalar values. The matrices _A_ and/or _B_ can optionally be transposed before performing the operation.", [ald_transa_m_k, bld_transb_k_n, cld_m]),
  Routine(True,  True,  0, False, "3", "symm",  T,  [S,D,C,Z,H],     ["m","n"],            ["layout","side","triangle"],                          ["a","b"],  ["c"],                        [ammn,bmnn,cmn], ["alpha","beta"], "",    "Symmetric matrix-matrix multiplication", "Same operation as xGEMM, but _A_ is symmetric instead. In case of `side == kLeft`, _A_ is a symmetric _m_ by _m_ matrix and _C = alpha * A * B + beta * C_ is performed. Otherwise, in case of `side == kRight`, _A_ is a symmtric _n_ by _n_ matrix and _C = alpha * B * A + beta * C_ is performed.", [ald_side_m_n, bld_m, cld_m]),
  Routine(True,  True,  0, False, "3", "hemm",  T,  [C,Z],           ["m","n"],            ["layout","side","triangle"],                          ["a","b"],  ["c"],                        [ammn,bmnn,cmn], ["alpha","beta"], "",    "Hermitian matrix-matrix multiplication", "Same operation as xSYMM, but _A_ is an Hermitian matrix instead.", [ald_side_m_n, bld_m, cld_m]),
  Routine(True,  True,  0, False, "3", "syrk",  T,  [S,D,C,Z,H],     ["n","k"],            ["layout","triangle","a_transpose"],                   ["a"],      ["c"],                        [ank,cn],        ["alpha","beta"], "",    "Rank-K update of a symmetric matrix", "Performs the matrix product _C = alpha * A * A^T + beta * C_ or _C = alpha * A^T * A + beta * C_, in which _A_ is a general matrix and _A^T_ is its transpose, _C_ (_n_ by _n_) is the symmetric matrix to be updated, and _alpha_ and _beta_ are scalar values.", [ald_trans_n_k, cld_m]),
  Routine(True,  True,  0, False, "3", "herk",  Tc, [Css,Zdd],       ["n","k"],            ["layout","triangle","a_transpose"],                   ["a"],      ["c"],                        [ank,cn],        ["alpha","beta"], "",    "Rank-K update of a hermitian matrix", "Same operation as xSYRK, but _C_ is an Hermitian matrix instead.", [ald_trans_n_k, cld_m]),
  Routine(True,  True,  0, False, "3", "syr2k", T,  [S,D,C,Z,H],     ["n","k"],            ["layout","triangle","ab_transpose"],                  ["a","b"],  ["c"],                        [ankab,bnkab,cn],["alpha","beta"], "",    "Rank-2K update of a symmetric matrix", "Performs the matrix product _C = alpha * A * B^T + alpha * B * A^T + beta * C_ or _C = alpha * A^T * B + alpha * B^T * A + beta * C_, in which _A_ and _B_ are general matrices and _A^T_ and _B^T_ are their transposed versions, _C_ (_n_ by _n_) is the symmetric matrix to be updated, and _alpha_ and _beta_ are scalar values.", [ald_trans_n_k, bld_trans_n_k, cld_n]),
  Routine(True,  True,  0, False, "3", "her2k", TU, [Ccs,Zzd],       ["n","k"],            ["layout","triangle","ab_transpose"],                  ["a","b"],  ["c"],                        [ankab,bnkab,cn],["alpha","beta"], "",    "Rank-2K update of a hermitian matrix", "Same operation as xSYR2K, but _C_ is an Hermitian matrix instead.", [ald_trans_n_k, bld_trans_n_k, cld_n]),
  Routine(True,  True,  0, False, "3", "trmm",  T,  [S,D,C,Z,H],     ["m","n"],            ["layout","side","triangle","a_transpose","diagonal"], ["a"],      ["b"],                        [amns,bmn],      ["alpha"],        "",    "Triangular matrix-matrix multiplication", "Performs the matrix product _B = alpha * A * B_ or _B = alpha * B * A_, in which _A_ is a unit or non-unit triangular matrix, _B_ (_m_ by _n_) is the general matrix to be updated, and _alpha_ is a scalar value.", [ald_side_m_n, bld_m]),
  Routine(True,  True,  0, False, "3", "trsm",  T,  [S,D,C,Z],       ["m","n"],            ["layout","side","triangle","a_transpose","diagonal"], ["a"],      ["b"],                        [amns,bmn],      ["alpha"],        "",    "Solves a triangular system of equations", "Solves the equation _A * X = alpha * B_ for the unknown _m_ by _n_ matrix X, in which _A_ is an _n_ by _n_ unit or non-unit triangular matrix and B is an _m_ by _n_ matrix. The matrix _B_ is overwritten by the solution _X_.", []),
],
[  # Level X: extra routines (not part of BLAS)
  # Special routines:
  Routine(True,  True,  0, False, "x", "had",      T, [S,D,C,Z,H],   ["n"],                [],                                                    ["x","y"],  ["z"],                        [xn,yn,zn],      ["alpha","beta"], "",    "Element-wise vector product (Hadamard)", "Performs the Hadamard element-wise product _z = alpha * x * y + beta * z_, in which _x_, _y_, and _z_ are vectors and _alpha_ and _beta_ are scalar constants.", []),
  Routine(True,  True,  0, False, "x", "omatcopy", T, [S,D,C,Z,H],   ["m","n"],            ["layout","a_transpose"],                              ["a"],      ["b"],                        [amn,bnma],      ["alpha"],        "",    "Scaling and out-place transpose/copy (non-BLAS function)", "Performs scaling and out-of-place transposition/copying of matrices according to _B = alpha*op(A)_, in which _A_ is an input matrix (_m_ rows by _n_ columns), _B_ an output matrix, and _alpha_ a scalar value. The operation _op_ can be a normal matrix copy, a transposition or a conjugate transposition.", [ald_m, bld_n]),
  Routine(True,  True,  0, False, "x", "im2col",   T, [S,D,C,Z,H],   im2col_constants,     ["kernel_mode"],                                       ["im"],     ["col"],                      [im,col],        [""],             "",    "Im2col function (non-BLAS function)", "Performs the im2col algorithm, in which _im_ is the input matrix and _col_ is the output matrix. Overwrites any existing values in the _col_ buffer", []),
  Routine(True,  True,  0, False, "x", "col2im",   T, [S,D,C,Z,H],   im2col_constants,     ["kernel_mode"],                                       ["col"],    ["im"],                       [col,im],        [""],             "",    "Col2im function (non-BLAS function)", "Performs the col2im algorithm, in which _col_ is the input matrix and _im_ is the output matrix. Accumulates results on top of the existing values in the _im_ buffer.", []),
  Routine(True,  True,  0, False, "x", "convgemm", T, [S,D,H],       convgemm_constants,   ["kernel_mode"],                                       ["im","kernel"], ["result"],              [imb,kernel,result],[""],          "",    "Batched convolution as GEMM (non-BLAS function)", "Integrates im2col and GEMM for batched 3D convolution, in which _im_ is the 4D input tensor (NCHW - batch-channelin-height-width), _kernel_ the 4D kernel weights tensor (KCHW - channelout-channelin-height-width), and _result_ the 4D output tensor (NCHW - batch-channelout-height-width).", []),
  # Batched routines:
  Routine(True,  True,  1, False, "x", "axpy",     T, [S,D,C,Z,H],   ["n"],                [],                                                    ["x"],      ["y"],                        [xn,yn],         ["alpha"],        "",    "Batched version of AXPY", "As AXPY, but multiple operations are batched together for better performance.", []),
  Routine(True,  True,  1, False, "x", "gemm",     T, [S,D,C,Z,H],   ["m","n","k"],        ["layout","a_transpose","b_transpose"],                ["a","b"],  ["c"],                        [amk,bkn,cmn],   ["alpha","beta"], "",    "Batched version of GEMM", "As GEMM, but multiple operations are batched together for better performance.", [ald_transa_m_k, bld_transb_k_n, cld_m]),
  Routine(True,  True,  2, False, "x", "gemm",     T, [S,D,C,Z,H],   ["m","n","k"],        ["layout","a_transpose","b_transpose"],                ["a","b"],  ["c"],                        [amk,bkn,cmn],   ["alpha","beta"], "",    "StridedBatched version of GEMM", "As GEMM, but multiple strided operations are batched together for better performance.", [ald_transa_m_k, bld_transb_k_n, cld_m]),
]]


def main(argv):

    # Parses the command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("clblast_root", help="Root of the CLBlast sources")
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase verbosity of the script")
    cl_args = parser.parse_args(argv)
    library_root = cl_args.clblast_root

    # Checks whether the command-line arguments are valid; exists otherwise
    for f in FILES:
        if not os.path.isfile(library_root + f):
            print("[ERROR] The path '" + library_root + "' does not point to the root of the CLBlast library")
            sys.exit()

    # Iterates over all regular files to output
    for i in range(0, len(FILES)):

        # Stores the header and the footer of the original file
        with open(library_root + FILES[i]) as f:
            original = f.readlines()
        file_header = original[:HEADER_LINES[i]]
        file_footer = original[-FOOTER_LINES[i]:]

        # Re-writes the body of the file
        with open(library_root + FILES[i], "w") as f:
            body = ""
            levels = [1, 2, 3] if (i == 4 or i == 5 or i == 6) else [1, 2, 3, 4]
            for level in levels:
                if i not in [11]:
                    body += cpp.LEVEL_SEPARATORS[level - 1] + "\n"
                for routine in ROUTINES[level - 1]:
                    if i == 0:
                        body += cpp.clblast_h(routine)
                    if i == 1:
                        body += cpp.clblast_cc(routine)
                    if i == 2:
                        body += cpp.clblast_c_h(routine)
                    if i == 3:
                        body += cpp.clblast_c_cc(routine)
                    if i == 4:
                        body += cpp.wrapper_clblas(routine)
                    if i == 5:
                        body += cpp.wrapper_cblas(routine)
                    if i == 6:
                        body += cpp.wrapper_cublas(routine)
                    if i == 7:
                        if routine.batched == 0 and routine.name not in ["convgemm"]:
                            body += cpp.clblast_netlib_c_h(routine)
                    if i == 8:
                        if routine.batched == 0 and routine.name not in ["convgemm"]:
                            body += cpp.clblast_netlib_c_cc(routine)
                    if i == 9:
                        body += cpp.clblast_h(routine, cuda=True)
                    if i == 10:
                        body += cpp.clblast_cc(routine, cuda=True)
                    if i == 11:
                        body += pyclblast.generate_pyx(routine)
            f.write("".join(file_header))
            f.write(body)
            f.write("".join(file_footer))

    # Outputs all the test implementations
    for level in [1, 2, 3, 4]:
        for routine in ROUTINES[level - 1]:
            if routine.has_tests:
                level_string = cpp.LEVEL_NAMES[level - 1]
                routine_suffix = "level" + level_string + "/x" + routine.lowercase_name() + ".cpp"

                # Correctness tests
                filename = library_root + "/test/correctness/routines/" + routine_suffix
                with open(filename, "w") as f:
                    f.write(cpp.HEADER + "\n")
                    f.write(cpp.correctness_test(routine, level_string))
                    f.write(cpp.FOOTER)

                # Performance tests
                filename = library_root + "/test/performance/routines/" + routine_suffix
                with open(filename, "w") as f:
                    f.write(cpp.HEADER + "\n")
                    f.write(cpp.performance_test(routine, level_string))
                    f.write(cpp.FOOTER)

    # API documentation
    filename = cl_args.clblast_root + "/doc/api.md"

    # Stores the header and the footer of the original documentation file
    with open(filename) as f:
        original = f.readlines()
    file_header = original[:HEADER_LINES_DOC]
    file_footer = original[-FOOTER_LINES_DOC:]

    # Outputs the API documentation
    with open(filename, "w") as f:

        # Outputs the header
        f.write("".join(file_header))
        doc_header = doc.header()
        f.write(doc_header)

        # Generates the documentation for each routine
        for level in [1, 2, 3, 4]:
            for routine in ROUTINES[level - 1]:
                if routine.implemented:
                    doc_routine = doc.generate(routine)
                    f.write(doc_routine)

        # Outputs the footer
        f.write("".join(file_footer))

if __name__ == '__main__':
    main(sys.argv[1:])



import _librapid


class Shape(_librapid.Shape):
    """
    Stores the dimensions of an N-dimensional Array.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
       


import os
import textwrap

import shapeGenerator
import arrayGenerator

outputDir = "../python/generated"

boilerplate = textwrap.dedent(f"""
            #pragma once

            #define LIBRAPID_ASSERT

            #include <librapid/librapid.hpp>
            #include <pybind11/pybind11.h>
            #include <pybind11/stl.h>
            #include <pybind11/functional.h>
            
            namespace py  = pybind11;
            namespace lrc = librapid;
        """).strip()


def main():
    # Ensure the output directory exists
    if not os.path.exists(outputDir):
        os.makedirs(outputDir)

    interfaceFunctions = []

    interfaceFunctions += shapeGenerator.write(outputDir)
    interfaceFunctions += arrayGenerator.write(outputDir)

    with open(f"{outputDir}/librapidPython.hpp", "w") as f:
        f.write(boilerplate)
        f.write("\n\n")
        for interfaceDef, _ in interfaceFunctions:
            f.write(f"{interfaceDef()};\n")

    with open(f"{outputDir}/librapidPython.cpp", "w") as f:
        f.write("#include \"librapidPython.hpp\"\n\n")
        f.write("PYBIND11_MODULE(_librapid, module) {\n")
        f.write("    module.doc() = \"Python bindings for librapid\";\n")
        for _, interfaceCall in interfaceFunctions:
            f.write(f"    {interfaceCall('module')};\n")
        f.write("}\n")


if __name__ == "__main__":
    main()



import argument
import function
import class_
import module
import file

methods = [
    # Default constructor
    function.Function(
        name="__init__",
        args=[]
    ),

    # List of values
    function.Function(
        name="__init__",
        args=[
            argument.Argument(
                name="vals",
                type="std::vector<uint32_t>",
                const=True,
                ref=True
            )
        ]
    ),

    # Copy constructor
    function.Function(
        name="__init__",
        args=[
            argument.Argument(
                name="other",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ]
    ),

    # zeros
    function.Function(
        name="zeros",
        args=[
            argument.Argument(
                name="dims",
                type="int",
                const=True,
                ref=False
            )
        ],
        static=True,
        op="""
            return lrc::Shape::zeros(dims);
        """
    ),

    # ones
    function.Function(
        name="ones",
        args=[
            argument.Argument(
                name="dims",
                type="int",
                const=True,
                ref=False
            )
        ],
        static=True,
        op="""
            return lrc::Shape::ones(dims);
        """
    ),

    # Indexing (get)
    function.Function(
        name="__getitem__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            ),
            argument.Argument(
                name="index",
                type="int",
                const=True,
                ref=False
            )
        ],
        op="""
            return self[index];
        """
    ),

    # Indexing (set)
    function.Function(
        name="__setitem__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=False,
                ref=True
            ),
            argument.Argument(
                name="index",
                type="int",
                const=True,
                ref=False
            ),
            argument.Argument(
                name="value",
                type="uint32_t",
                const=True,
                ref=False
            )
        ],
        op="""
            self[index] = value;
        """,
        isOperator=True
    ),

    # Equality
    function.Function(
        name="__eq__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            ),
            argument.Argument(
                name="other",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return self == other;
        """,
        isOperator=True
    ),

    # Inequality
    function.Function(
        name="__ne__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            ),
            argument.Argument(
                name="other",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return self != other;
        """,
        isOperator=True
    ),

    # Number of dimensions
    function.Function(
        name="ndim",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return self.ndim();
        """
    ),

    # Subshape
    function.Function(
        name="subshape",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            ),
            argument.Argument(
                name="start",
                type="int",
                const=True,
                ref=False
            ),
            argument.Argument(
                name="end",
                type="int",
                const=True,
                ref=False
            )
        ],
        op="""
            return self.subshape(start, end);
        """
    ),

    # Size
    function.Function(
        name="size",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return self.size();
        """
    ),

    # To string (__str__)
    function.Function(
        name="__str__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return fmt::format("{}", self);
        """
    ),

    # To string (__repr__)
    function.Function(
        name="__repr__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            )
        ],
        op="""
            return fmt::format("_librapid.{}", self);
        """
    ),

    # Format (__format__)
    function.Function(
        name="__format__",
        args=[
            argument.Argument(
                name="self",
                type="lrc::Shape",
                const=True,
                ref=True
            ),
            argument.Argument(
                name="formatSpec",
                type="std::string",
                const=True,
                ref=True
            )
        ],
        op="""
            std::string format = fmt::format("{{:{}}}", formatSpec);
            return fmt::format(fmt::runtime(format), self);
        """
    )
]

classType = class_.Class(
    name="Shape",
    type="lrc::Shape",
)

classType.functions = methods

moduleType = module.Module(
    name="librapid.shape"
)

moduleType.classes.append(classType)


def write(root):
    fileType = file.File(
        path=f"{root}/shape.cpp"
    )

    fileType.modules.append(moduleType)

    interfaceFunctions = fileType.write()
    # Run clang-format if possible
    try:
        import subprocess

        subprocess.run(["clang-format", "-i", fileType.path])
    except Exception as e:
        print("Unable to run clang-format:", e)

    return interfaceFunctions



from class_ import Class
import textwrap


class Module:
    def __init__(self, name, parentModule=None, docstring=None, includeGuard=None):
        self.name = name
        self.parent = parentModule
        self.docstring = docstring
        self.includeGuard = includeGuard
        self.classes = []
        self.functions = []

    def addClass(self, class_):
        self.classes.append(class_)
        return self

    def addFunction(self, func):
        self.functions.append(func)
        return self

    def genInterfaceDefinition(self):
        tmpName = self.name.replace(".", "_")
        return f"void genInterface_{tmpName}(py::module& module)"

    def genInterfaceCall(self, moduleName):
        tmpName = self.name.replace(".", "_")
        return f"genInterface_{tmpName}({moduleName})"

    def genInterface(self):
        ret = f"{self.genInterfaceDefinition()} {{\n"

        if self.parent is None:
            moduleName = "module"
        else:
            ret += f"module.def_submodule(\"{self.name}\", \"{self.parent.name}.{self.name}\") {{\n"
            moduleName = self.name

        if self.docstring is not None:
            ret += f"{moduleName}.doc() = \"{self.docstring}\";\n\n"

        for class_ in self.classes:
            ret += class_.genInterface(moduleName)
            ret += "\n"

        for func in self.functions:
            ret += func.gen(moduleName)
            ret += "\n"

        ret += "}\n"

        if self.includeGuard is None:
            return ret
        else:
            return textwrap.dedent(f"""
            #if {self.includeGuard}
            {ret}
            #else
            {self.genInterfaceDefinition()} {{}}
            #endif
            """)


if __name__ == "__main__":
    testModule = Module("test")



import itertools


class Swizzler:
    def __init__(self, swiz, vecName="GenericVector"):
        self.swiz = "".join(swiz)
        self.vec_name = vecName

    def getter_definition(self):
        # "LIBRAPID_NODISCARD LIBRAPID_ALWAYS_INLINE GenericVector<Scalar, 2> xy() const;"
        return f"LIBRAPID_NODISCARD LIBRAPID_ALWAYS_INLINE {self.vec_name}<Scalar, {len(self.swiz)}> {self.swiz}() const;"

    def setter_definition(self):
        # "LIBRAPID_ALWAYS_INLINE void xy(const GenericVector<Scalar, 2> &v);"
        return f"LIBRAPID_ALWAYS_INLINE void {self.swiz}(const {self.vec_name}<Scalar, {len(self.swiz)}> &v);"

    def getter_implementation(self):
        """
        template<typename Scalar, int64_t Dims>
        auto GenericVector<Scalar, Dims>::xy() const -> GenericVector<Scalar, 2> {
            return {x(), y()};
        }
        """

        params = ", ".join([x + "()" for x in self.swiz])
        return f"template<typename Scalar, int64_t Dims>\nauto {self.vec_name}<Scalar, Dims>::{self.swiz}() const -> {self.vec_name}<Scalar, {len(self.swiz)}> {{\n    return {{{params}}};\n}}"

    def setter_implementation(self):
        """
        template<typename Scalar, int64_t Dims>
        void GenericVector<Scalar, Dims>::xy(const GenericVector<Scalar, 2> &v) {
            x(v.x());
            y(v.y());
        }
        """

        assignments = ""
        for a, b in zip(self.swiz, list("xyzw")):
            assignments += f"{a}(v.{b}());\n"
        return f"template<typename Scalar, int64_t Dims>\nvoid {self.vec_name}<Scalar, Dims>::{self.swiz}(const {self.vec_name}<Scalar, {len(self.swiz)}> &v) {{\n{assignments}}}"

    def test_getter(self):
        """
        REQUIRE(testC.xy() == lrc::Vec2d(1, 2));
        """
        tmpArgs = {"x": "1", "y": "2", "z": "3", "w": "4"}
        return f"REQUIRE(testC.{self.swiz}() == lrc::Vec{len(self.swiz)}d({', '.join([tmpArgs[v] for v in self.swiz])}));"

    def test_setter(self):
        """
        testC.xy(lrc::Vec2d(1, 2));
        REQUIRE(testC.x() == 1);
        REQUIRE(testC.y() == 2);
        """
        tmpArgs = {"x": "1", "y": "2", "z": "3", "w": "4"}
        return f"testC.{self.swiz}(lrc::Vec{len(self.swiz)}d({', '.join([tmpArgs[v] for v in self.swiz])}));\n" + "\n".join(
            [f"REQUIRE(testC.{v}() == {tmpArgs[v]});" for v in self.swiz])


to_swizzle = ["xy", "xz", "yz", "xyz", "xyw", "xzw", "yzw", "xyzw"]

vecName = "GenericVector"

getter_implementations = []
setter_implementations = []
getter_definitions = []
setter_definitions = []
setters = []
for swiz in to_swizzle:
    for perm in itertools.permutations(list(swiz)):
        swizzler = Swizzler(perm, vecName)
        getter_definitions.append(swizzler.getter_definition())
        setter_definitions.append(swizzler.setter_definition())
        getter_implementations.append(swizzler.getter_implementation())
        setter_implementations.append(swizzler.setter_implementation())

print("============== GETTER DEFINITIONS ==============")
print("\n".join(getter_definitions))
print("============== SETTER DEFINITIONS ==============")
print("\n".join(setter_definitions))
print("============== GETTER IMPLEMENTATIONS ==============")
print("\n\n".join(getter_implementations))
print("============== SETTER IMPLEMENTATIONS ==============")
print("\n\n".join(setter_implementations))

# for swiz in to_swizzle:
#     for perm in itertools.permutations(list(swiz)):
#         term = ", ".join([x + "()" for x in perm])
#         print(
#             f"""LIBRAPID_NODISCARD LIBRAPID_ALWAYS_INLINE {vecName}<Scalar, {len(perm)}> {"".join(perm)}() const {{ return {{ {term} }}; }}""")
#
#         assignments = ""
#         for a, b in zip(perm, list("xyzw")):
#             assignments += f"{a}(vec.{b}());\n"
#         print(
#             f"""template<typename S, size_t D> LIBRAPID_ALWAYS_INLINE void {"".join(perm)}(const {vecName}<S, D> &vec) const {{ {assignments} }}""")

# tmpArgs = {"x": "1", "y": "2", "z": "3", "w": "4"}
# for swiz in to_swizzle:
#     for perm in itertools.permutations(list(swiz)):
#         term = ", ".join([x + "()" for x in perm])
#         print(f"""REQUIRE(testC.{"".join(perm)}() == lrc::Vec{len(perm)}d({", ".join([tmpArgs[v] for v in perm])}));""")




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0.
# This file follows the PEP8 Python style guide and uses a max-width of 100 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

from setuptools import setup

from distutils.extension import Extension
from Cython.Distutils import build_ext
import platform

runtime_library_dirs = list()
if platform.system() == "Linux":
    runtime_library_dirs.append("/usr/local/lib")
elif platform.system() == "Windows":
    runtime_library_dirs.append("C:/Program Files/clblast/lib")
    runtime_library_dirs.append("C:/Program Files (x86)/clblast/lib")

ext_modules = list()
ext_modules.append(
    Extension(
        "pyclblast",
        ["src/pyclblast.pyx"],
        libraries=["clblast"],
        runtime_library_dirs=runtime_library_dirs,
        language="c++"
    )
)

setup(
    name="pyclblast",
    version="1.3.2",
    author="Cedric Nugteren",
    author_email="web@cedricnugteren.nl",
    url="https://github.com/CNugteren/CLBlast/blob/master/src/pyclblast",
    description="Python bindings for CLBlast, the tuned OpenCL BLAS library",
    license="Apache Software License",
    requires=["numpy", "pyopencl", "cython"],
    package_dir={'': 'src'},
    scripts=[],
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords="OpenCL BLAS CLBlast GEMM matrix-multiplication"
)



DTYPES = [
    # {
    #     "name": "char",
    #     "cname": "char",
    #     "floating": False
    # },
    # {
    #     "name": "uchar",
    #     "cname": "unsigned char",
    #     "floating": False
    # },
    # {
    #     "name": "short",
    #     "cname": "short",
    #     "floating": False
    # },
    # {
    #     "name": "ushort",
    #     "cname": "unsigned short",
    #     "floating": False
    # },
    # {
    #     "name": "int",
    #     "cname": "int",
    #     "floating": False
    # },
    # {
    #     "name": "uint",
    #     "cname": "unsigned int",
    #     "floating": False
    # },
    # {
    #     "name": "long",
    #     "cname": "long",
    #     "floating": False
    # },
    # {
    #     "name": "ulong",
    #     "cname": "unsigned long",
    #     "floating": False
    # },
    # {
    #     "name": "longlong",
    #     "cname": "long long",
    #     "floating": False
    # },
    # {
    #     "name": "ulonglong",
    #     "cname": "unsigned long long",
    #     "floating": False
    # },
    {
        "name": "F",
        "cname": "FL",
        "floating": True
    },
    {
        "name": "D",
        "cname": "DL",
        "floating": True
    }
]
FLOATING = ["float", "double"]
MIN_DIMS = 2
MAX_DIMS = 4

FUNCS = [
    {
        "name": "+",
        "op": "+",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "-",
        "op": "-",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "*",
        "op": "*",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "/",
        "op": "/",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "+=",
        "op": "+",
        "opType": "binary",
        "inplace": True,
        "requireFloating": False
    },
    {
        "name": "-=",
        "op": "-",
        "opType": "binary",
        "inplace": True,
        "requireFloating": False
    },
    {
        "name": "*=",
        "op": "*",
        "opType": "binary",
        "inplace": True,
        "requireFloating": False
    },
    {
        "name": "/=",
        "op": "/",
        "opType": "binary",
        "inplace": True,
        "requireFloating": False
    },
    {
        "name": ">",
        "op": ">",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "<",
        "op": "<",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": ">=",
        "op": ">=",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "<=",
        "op": "<=",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "==",
        "op": "==",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "!=",
        "op": "!=",
        "opType": "binary",
        "inplace": False,
        "requireFloating": False
    },
    {
        "name": "sin",
        "op": "sin",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "cos",
        "op": "cos",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "tan",
        "op": "tan",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "asin",
        "op": "asin",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "acos",
        "op": "acos",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "atan",
        "op": "atan",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "sinh",
        "op": "sinh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "cosh",
        "op": "cosh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "tanh",
        "op": "tanh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "asinh",
        "op": "asinh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "acosh",
        "op": "acosh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "atanh",
        "op": "atanh",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "exp",
        "op": "exp",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "log",
        "op": "log",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "log2",
        "op": "log2",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "log10",
        "op": "log10",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "sqrt",
        "op": "sqrt",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "cbrt",
        "op": "cbrt",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "abs",
        "op": "abs",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "ceil",
        "op": "ceil",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    },
    {
        "name": "floor",
        "op": "floor",
        "opType": "unary",
        "inplace": False,
        "requireFloating": True
    }
]

def gen():
    ret = ""
    for type in DTYPES:
        for func in FUNCS:
            if func["requireFloating"] and not type["floating"]:
                continue

            for dims in range(MIN_DIMS, MAX_DIMS + 1):
                if func["opType"] == "binary":
                    if func["inplace"]:
                        op = func["op"]
                        block = f"""IOF({type["name"]}{dims},{func["name"]})({type["name"]}{dims} &a,CO {type["name"]}{dims} &b){{M{type["name"]}{dims}N({",".join([f"a.{var}{op}b.{var}" for var in "xyzw"[:dims]])});}}
IOF({type["name"]}{dims},{func["name"]})({type["name"]}{dims} &a,CO {type["cname"]} &b){{M{type["name"]}{dims}N({",".join([f"a.{var} {op} b" for var in "xyzw"[:dims]])});}}
"""

                        ret += block
                    else:
                        op = func["op"]
                        block = f"""IOF({type["name"]}{dims},{func["name"]})(CO {type["name"]}{dims} &a,CO {type["name"]}{dims} &b){{M{type["name"]}{dims}({",".join([f"a.{var}{op}b.{var}" for var in "xyzw"[:dims]])});}}
IOF({type["name"]}{dims},{func["name"]})(CO {type["name"]}{dims} &a,CO {type["cname"]} &b){{M{type["name"]}{dims}({",".join([f"a.{var}{op}b" for var in "xyzw"[:dims]])});}}
IOF({type["name"]}{dims},{func["name"]})(CO {type["cname"]} &a,CO {type["name"]}{dims} &b){{M{type["name"]}{dims}({",".join([f"a{op}b.{var}" for var in "xyzw"[:dims]])});}}
"""

                        ret += block
                elif func["opType"] == "unary":
                    op = func["op"]
                    block = f"""inline {type["name"]}{dims} {func["name"]}(CO {type["name"]}{dims} &a){{M{type["name"]}{dims}({",".join([f"{op}(a.{var})" for var in "xyzw"[:dims]])});}}"""
                    ret += block + "\n"

    return ret

print(gen())



import textwrap


class File:
    def __init__(self, path=None, docstring=None):
        self.path = path if path is not None else "./module.cpp"
        self.modules = []

    def addModule(self, module):
        self.modules.append(module)
        return self

    def genInterface(self):
        interfaceFunctions = []
        ret = ""
        for module in self.modules:
            ret += module.genInterface()
            ret += "\n"

            interfaceFunctions.append((module.genInterfaceDefinition, module.genInterfaceCall))

        return ret, interfaceFunctions

    def write(self, path=None):
        interfaceFunctions = []
        with open(path if path is not None else self.path, "w") as f:
            f.write("#include \"librapidPython.hpp\"\n\n")
            interface, interfaceFunctionsTmp = self.genInterface()
            f.write(interface.strip())
            interfaceFunctions.extend(interfaceFunctionsTmp)

        return interfaceFunctions



from argument import Argument
import textwrap

RETURN_TAKE_OWNERSHIP = "take_ownership"
RETURN_COPY = "copy"
RETURN_MOVE = "move"
RETURN_REFERENCE = "reference"
RETURN_REFERENCE_INTERNAL = "reference_internal"
RETURN_AUTO = "automatic"
RETURN_AUTO_REFERENCE = "automatic_reference"


class Function:
    def __init__(self, *args, **kwargs):
        """
        Arguments in order:
        - name
        - args
        - op
        - static
        - property
        - returnPolicy
        - isOperator

        :param args:
        :param kwargs:
        """

        self.name = kwargs.get("name", None)
        self.args = kwargs.get("args", [])
        self.op = kwargs.get("op", None)
        self.static = kwargs.get("static", False)
        self.property = kwargs.get("property", False)
        self.returnPolicy = kwargs.get("returnPolicy", None)
        self.isOperator = kwargs.get("isOperator", False)

        for i in range(len(args)):
            if i == 0 and self.name is None:
                self.name = args[i]
            elif i == 1 and self.args is None:
                self.args = args[i]
            elif i == 2 and self.op is None:
                self.op = args[i]
            elif i == 3 and self.static is None:
                self.static = args[i]
            elif i == 4 and self.property is None:
                self.property = args[i]
            elif i == 5 and self.returnPolicy is None:
                self.returnPolicy = args[i]
            elif i == 6 and self.isOperator is None:
                self.isOperator = args[i]
            else:
                raise ValueError("Too many arguments")

        if self.name is None:
            raise ValueError("Function must have a name")

        self.isConstructor = self.name == "__init__"

        if self.op is None:
            if not self.isConstructor:
                raise ValueError("A non-constructor function must have an operation")
        else:
            self.op = textwrap.dedent(self.op).strip()

        if not isinstance(self.args, (list, tuple)):
            self.args = [self.args]

    def arguments(self, ):
        for arg in self.args:
            if isinstance(arg, Argument):
                yield arg
            else:
                yield Argument(arg)

    def genArgumentStr(self):
        return ", ".join([arg.param() for arg in self.arguments()])

    def pyargs(self):
        args = [arg.pyarg() for arg in self.arguments() if arg.name not in ["self", "*args", "**kwargs"]]
        if self.returnPolicy is not None:
            args.append(f"py::return_value_policy::{self.returnPolicy}")

        if self.isOperator:
            args.append("py::is_operator()")

        return ", ".join(args)

    def gen(self, parent, haveParent=True):
        defType = "def"
        if self.static:
            defType = "def_static"
        elif self.property:
            defType = "def_property"

        pyArgStr = self.pyargs()
        if pyArgStr != "":
            pyArgStr = ", " + pyArgStr

        # Special case for constructors
        if self.isConstructor and self.op is None:
            return textwrap.dedent(f"""
                {parent if haveParent else ""}.def(py::init([]({", ".join([arg.param() for arg in self.arguments()])}) {{
                    return {parent.type}({", ".join([arg.name for arg in self.arguments()])});
                }}){pyArgStr})
            """).strip()

        return textwrap.dedent(f"""
            {parent if haveParent else ""}.{defType}(\"{self.name}\", []({self.genArgumentStr()}) {{
                {self.op}
            }}{pyArgStr})
        """).strip()

    def __str__(self):
        return self.gen("module")


if __name__ == "__main__":
    testFunction = Function(
        name="test",
        args=[
            Argument("a", "int"),
            Argument("b", "int", default=0),
        ],
        op="""
            return a + b;
        """
    )

    print(testFunction.genArgumentStr())
    print(testFunction.pyargs())

    print(testFunction.gen("module"))



from argument import Argument
import function


class Class:
    def __init__(self, name, type):
        self.name = name
        self.type = type
        self.functions = []
        self.implicitConversions = []

    def addFunction(self, func):
        self.functions.append(func)
        return self

    def addImplicitConversion(self, other):
        self.implicitConversions.append(other)
        return self

    def genImplicitConversions(self):
        ret = ""
        for other in self.implicitConversions:
            ret += f"py::implicitly_convertible<{other.type}, {self.type}>();\n"

        return ret

    def genInterface(self, parent="module"):
        ret = f"py::class_<{self.type}>({parent}, \"{self.name}\")\n"
        for func in self.functions:
            ret += func.gen(self, False)

            if func is not self.functions[-1]:
                ret += "\n"

        ret += ";\n"

        if len(self.implicitConversions) > 0:
            ret += "\n"
            ret += self.genImplicitConversions()

        return ret

    def __str__(self):
        return self.name


if __name__ == "__main__":
    vector = Class("Vec2d", "lrc::Vec2d")
    vector.addFunction(function.Function(
        name="__init__",
        args=[]
    ))
    vector.addFunction(function.Function(
        name="__init__",
        args=[
            Argument("x", "double"),
            Argument("y", "double", default=0)
        ]
    ))
    vector.addFunction(function.Function(
        name="__init__",
        args=[
            Argument(
                name="other",
                type="lrc::Vec2d",
                const=True,
                ref=True
            )
        ]
    ))

    vector.addFunction(function.Function(
        name="__add__",
        args=[
            Argument(
                name="self",
                type="lrc::Vec2d",
                const=True,
                ref=True
            ),
            Argument(
                name="other",
                type="lrc::Vec2d",
                const=True,
                ref=True
            )
        ],
        op="""
            return (self + other).eval();
        """,
        isOperator=True
    ))

    vector.addFunction(function.Function(
        name="__str__",
        args=[
            Argument(
                name="self",
                type="lrc::Vec2d",
                const=True,
                ref=True
            )
        ],
        op="""
            return fmt::format("{}", self);
        """,
        isOperator=True
    ))

    vector.addImplicitConversion(vector)

    print(vector.genInterface())



# Configuration file for the Sphinx documentation builder

import os
import sys

import regex

sys.path.insert(0, os.path.abspath("../.."))

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_dirs = [
    "librapid/vendor",
    "librapid/blas",
    "librapid/cxxblas",
    "librapid/fftw",
    "librapid/cuda/kernels",
    "librapid/opencl/kernels",
    "librapid/opencl/opencl.hpp"
]

file_match = regex.compile(".*\..*")

# -- Project information -----------------------------------------------------

project = "librapid"
copyright = "2023, Toby Davis"
author = "Toby Davis"

# The full version, including alpha/beta/rc tags
currentMajorVersion = None
currentMinorVersion = None
currentPatchVersion = None

try:
    with open("../../version.txt") as versionFile:
        text = versionFile.read()
        currentMajorVersion = regex.search("MAJOR [0-9]+", text).group().split()[1]
        currentMinorVersion = regex.search("MINOR [0-9]+", text).group().split()[1]
        currentPatchVersion = regex.search("PATCH [0-9]+", text).group().split()[1]
    print(f"Current Version: v{currentMajorVersion}.{currentMinorVersion}.{currentPatchVersion}")
except Exception as e:
    print("[ ERROR ] Failed to read version.txt")
    print(e)
    sys.exit(1)

release = f"v{currentMajorVersion}.{currentMinorVersion}.{currentPatchVersion}"

# Set the master file
master_doc = "index"

# -- Run Doxygen -------------------------------------------------------------

try:
    # subprocess.run(["cd ../..", "doxygen html latex xml"])
    os.system("cd ../.. && doxygen")
except Exception as e:
    print("[ ERROR ] Failed to run doxygen")
    print(e)
    print("\n\nPlease run Doxygen yourself from the source directory")
    sys.exit(1)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "breathe",
    "exhale",
    "numpydoc",
    "sphinx_favicon",
    "myst_parser",
    "sphinx_design",
    "sphinx_copybutton"
]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
]

# Download the benchmark results from GitHub Actions
if os.environ.get("GITHUB_TOKEN") is not None:
    extensions.append("rtds_action")
    rtds_action_github_repo = "LibRapid/librapid"
    rtds_action_path = "BenchmarkResults"
    rtds_action_artifact_prefix = "LibRapid_Benchmark_SHA_"
    rtds_action_github_token = os.environ["GITHUB_TOKEN"]
    rtds_action_error_if_missing = False # Don't fail if the artifact is missing

autosectionlabel_prefix_document = True

# Set up the breathe extension
breathe_projects = {
    "librapid": "../xml"
}

breathe_default_project = "librapid"

# # Custom inputs to doxygen generator
# doxygen_inputs = """
#
# INPUT				 = ./librapid_doc_copy
#
# ENABLE_PREPROCESSING  = YES
# MACRO_EXPANSION	   = YES
# EXPAND_ONLY_PREDEF	= NO
# PREDEFINED			+= LIBRAPID_DOXYGEN_BUILD
# PREDEFINED			+= LR_INLINE=
# PREDEFINED			+= __restrict=
# PREDEFINED			+= LIBRAPID_MAX_DIMS=32
# PREDEFINED			+= __host__=
# PREDEFINED			+= __device__=
# PREDEFINED			+= __global__=
#
# """

# Set up the exhale extension
exhale_args = {
    # These arguments are required
    "containmentFolder": "./api",
    "rootFileName": "index.rst",
    "rootFileTitle": "LibRapid",
    "createTreeView": True,
    "doxygenStripFromPath": "../..",
    "verboseBuild": True
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "*jitify*",
    "*fmt*",
    "*blas/*",
    "*bindings*"
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_title = "LibRapid Docs"
html_theme = "furo"
# html_theme = "pydata_sphinx_theme"
# html_theme = "sphinx_book_theme"

html_logo = "../../branding/LibRapid small space.png"

html_context = {
    "display_github": True,
    "github_user": "LibRapid",
    "github_repo": "librapid",
    "github_version": "master",
    "doc_path": "docs/source",
    "conf_py_path": "docs/source"
}

html_theme_options = {
    # Show a banner at the top of the page
    "announcement": """
<div style="white-space = word-break!important">
    <img src="/en/latest/_static/LR_icon.png" alt="LibRapid" width="22.5"></img>
    <a href="https://github.com/sponsors/Pencilcaseman">
        <span class="banner-resize-text-lg">
            If you like LibRapid, please consider supporting its development!
        </span>
        <span class="banner-resize-text-sm">
            Support LibRapid's Development!
        </span>
    </a>
</div>
	""",
    "sidebar_hide_name": True,
    "light_logo": "LibRapid_light.png",
    "dark_logo": "LibRapid_dark.png",
}

# CopyButton configuration
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"
copybutton_here_doc_delimiter = "EOT"
copybutton_selector = "div:not(.no-copybutton) > div.highlight > pre"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]
html_css_files = ["css/s4defs-roles.css", "css/custom.css"]

# html_favicon = "/en/latest/_static/LR_icon_128.png"
favicons = [
    {
        "rel": "icon",
        "static-file": "/en/latest/_static/LR_icon_128.png",
        "type": "image/png",
    },
    {
        "rel": "icon",
        "sizes": "16x16",
        "href": "/en/latest/_static/LR_icon_128.png",
        "type": "image/png",
    },
    {
        "rel": "icon",
        "sizes": "32x32",
        "href": "/en/latest/_static/LR_icon_128.png",
        "type": "image/png",
    },
    {
        "rel": "icon",
        "sizes": "64x64",
        "href": "/en/latest/_static/LR_icon_128.png",
        "type": "image/png",
    },
    {
        "rel": "apple-touch-icon",
        "sizes": "180x180",
        "href": "/en/latest/_static/LR_icon.png",
        "type": "image/png",
    },
]

rst_prolog = """
.. include:: <s5defs.txt>
"""

# Set LaTeX preamble to allow for deeper nesting of itemize environments
latex_elements = {
    "preamble": r"""
\usepackage{enumitem}
\setlistdepth{9}
\setlist[itemize,1]{label=\textbullet}
\setlist[itemize,2]{label=--}
\setlist[itemize,3]{label=*}
\setlist[itemize,4]{label=-}
\setlist[itemize,5]{label=$\circ$}
\setlist[itemize,6]{label=$\bullet$}
\setlist[itemize,7]{label=$\diamond$}
\setlist[itemize,8]{label=$\ast$}
\setlist[itemize,9]{label=$\cdot$}
\renewlist{itemize}{itemize}{9}
  """
}



import sys
"""

This script accepts a version number in the form X.Y.Z, where X, Y, and Z are integers.
It will then update all LibRapid version numbers to the specified version.

"""

import sys
import re
import argparse
from datetime import datetime

# Extract current version number
currentMajorVersion = None
currentMinorVersion = None
currentPatchVersion = None

try:
    with open("../version.txt", "r") as versionFile:
        text = versionFile.read()
        currentMajorVersion = re.search("MAJOR [0-9]+", text).group().split()[1]
        currentMinorVersion = re.search("MINOR [0-9]+", text).group().split()[1]
        currentPatchVersion = re.search("PATCH [0-9]+", text).group().split()[1]
    print(f"Current Version: v{currentMajorVersion}.{currentMinorVersion}.{currentPatchVersion}")
except Exception as e:
    print("[ ERROR ] Failed to read version.txt")
    print(e)
    sys.exit(1)

argParser = argparse.ArgumentParser()
argParser.add_argument("-v", "--version", type=str, help="Full Version")
argParser.add_argument("-M", "--major", type=int, help="Major Version")
argParser.add_argument("-m", "--minor", type=int, help="Minor Version")
argParser.add_argument("-p", "--patch", type=int, help="Patch Version")

args = argParser.parse_args()

if args.version and any([args.major, args.minor, args.patch]):
    print("[ ERROR ] -v and -M options cannot be used together")
    sys.exit(1)

if args.version:
    # Validate version number
    if not re.match("[0-9]+\\.[0-9]+\\.[0-9]+", args.version):
        print("[ ERROR ] Invalid version number")
        sys.exit(1)
    newMajorVersion = args.version.split(".")[0]
    newMinorVersion = args.version.split(".")[1]
    newPatchVersion = args.version.split(".")[2]
else:
    newMajorVersion = args.major if args.major else currentMajorVersion
    newMinorVersion = args.minor if args.minor else currentMinorVersion
    newPatchVersion = args.patch if args.patch else str(int(currentPatchVersion) + 1)
    
print(f"New Version: v{newMajorVersion}.{newMinorVersion}.{newPatchVersion}\n\n")

# Write to version.txt
with open("../version.txt", "w") as versionFile:
    versionFile.write(f"MAJOR {newMajorVersion}\n")
    versionFile.write(f"MINOR {newMinorVersion}\n")
    versionFile.write(f"PATCH {newPatchVersion}\n")
    print("Written to version.txt")

# Write to Doxyfile
template = None
with open("tmp/doxyTemplate", "r") as templateFile:
    template = templateFile.read()
    print("Loaded Doxyfile template")

with open("../Doxyfile", "w") as doxyfile:
    versionString = f"PROJECT_NUMBER         = v{newMajorVersion}.{newMinorVersion}.{newPatchVersion}"
    template = template.replace("$${{ INSERT_VERSION_NUMBER_HERE }}$$", versionString)
    doxyfile.write(template)
    print("Written to Doxyfile")

# Write to CITATION.cff
with open("tmp/citationTemplate.cff", "r") as templateFile:
    template = templateFile.read()
    print("Loaded CITATION.cff template")

with open("../CITATION.cff", "w") as citationFile:
    versionString = f"version: {newMajorVersion}.{newMinorVersion}.{newPatchVersion}"
    dateString = f"date-released: \"{datetime.now().strftime('%Y-%m-%d')}\""
    template = template.replace("$${{ INSERT_VERSION_NUMBER_HERE }}$$", versionString)
    template = template.replace("$${{ INSERT_DATE_HERE }}$$", dateString)
    citationFile.write(template)
    print("Written to CITATION.cff")

# Write to pyproject.toml
with open("tmp/pyprojectTemplate.toml", "r") as templateFile:
    template = templateFile.read()
    print("Loaded pyproject.toml template")

with open("../pyproject.toml", "w") as pyprojectFile:
    versionString = f"{newMajorVersion}.{newMinorVersion}.{newPatchVersion}"
    template = template.replace("$${{ INSERT_VERSION_NUMBER_HERE }}$$", versionString)
    pyprojectFile.write(template)
    print("Written to pyproject.toml")



import re


def isPrimitive(type):
    if type in ["bool", "int", "char", "float", "double"]:
        return True

    # Match int, int32_t, int64_t, uint32_t etc.
    if re.match(r"u?int\d+_t", type):
        return True

    return False


class Argument:
    def __init__(self, *args, **kwargs):
        """
        Arguments may include (in order):
        - name
        - type
        - default
        - const
        - ref
        - pointer
        - noConvert
        - returnPolicy

        :param args:
        :param kwargs (optional):
        """

        self.name = kwargs.get("name", None)
        self.type = kwargs.get("type", None)
        self.default = kwargs.get("default", None)
        self.const = kwargs.get("const", True)
        self.ref = kwargs.get("ref", True)
        self.pointer = kwargs.get("pointer", False)
        self.noConvert = kwargs.get("noConvert", False)

        for i in range(len(args)):
            if i == 0 and self.name is None:
                self.name = args[i]
            elif i == 1 and self.type is None:
                self.type = args[i]
            elif i == 2 and self.default is None:
                self.default = args[i]
            elif i == 3 and self.const is None:
                self.const = args[i]
            elif i == 4 and self.ref is None:
                self.ref = args[i]
            elif i == 5 and self.pointer is None:
                self.pointer = args[i]
            elif i == 6 and self.noConvert is None:
                self.noConvert = args[i]
            else:
                raise ValueError("Too many arguments")

        if self.name is None:
            raise ValueError("Argument must have a name")

        if self.type is None:
            raise ValueError("Argument must have a type")

        self.isArgs = self.name == "*args"
        self.isKwargs = self.name == "**kwargs"

    def param(self):
        if self.isArgs:
            return f"py::args args"
        elif self.isKwargs:
            return f"py::kwargs kwargs"
        else:
            isPrimitiveType = isPrimitive(self.type)
            return f"{'const ' if self.const and not isPrimitiveType else ''}{self.type} {'&' if self.ref and not isPrimitiveType else ''}{'*' if self.pointer else ''}{self.name}"

    def declaration(self):
        if self.default is None:
            return f"{self.type} {self.name}"
        else:
            return f"{self.type} {self.name} = {self.default}"

    def pyarg(self):
        if self.default is not None:
            return f"py::arg_v(\"{self.name}\", {self.default}, \"{self.name} = {self.default}\"){'.noconvert()' if self.noConvert else ''}"
        else:
            return f"py::arg(\"{self.name}\"){'.noconvert()' if self.noConvert else ''}"

    def __str__(self):
        return self.name


if __name__ == "__main__":
    # Run some tests

    normalArgNoDefault1 = Argument("arg1", "int")
    normalArgNoDefault2 = Argument(name="arg1", type="int")

    normalArgDefault1 = Argument("arg1", "int", 0)
    normalArgDefault2 = Argument(name="arg1", type="int", default=0)

    constArgNoDefault1 = Argument("arg1", "int", const=True)
    constArgNoDefault2 = Argument(name="arg1", type="int", const=True)

    refArgNoDefault1 = Argument("arg1", "int", ref=True, noConvert=True)
    refArgNoDefault2 = Argument(name="arg1", type="int", ref=True, noConvert=True)

    noConvertArgDefault = Argument(name="arg1", type="int", default=0, noConvert=True)

    print(f"Normal Argument ~ No Default ~ 1 ~ Param: {normalArgNoDefault1.param()}")
    print(f"Normal Argument ~ No Default ~ 2 ~ Param: {normalArgNoDefault2.param()}")
    print(f"Normal Argument ~ No Default ~ 1 ~ Declaration: {normalArgNoDefault1.declaration()}")
    print(f"Normal Argument ~ No Default ~ 2 ~ Declaration: {normalArgNoDefault2.declaration()}")
    print(f"Normal Argument ~ No Default ~ 1 ~ PyArg: {normalArgNoDefault1.pyarg()}")
    print(f"Normal Argument ~ No Default ~ 2 ~ PyArg: {normalArgNoDefault2.pyarg()}")

    print(f"Normal Argument ~ Default ~ 1 ~ Param: {normalArgDefault1.param()}")
    print(f"Normal Argument ~ Default ~ 2 ~ Param: {normalArgDefault2.param()}")
    print(f"Normal Argument ~ Default ~ 1 ~ Declaration: {normalArgDefault1.declaration()}")
    print(f"Normal Argument ~ Default ~ 2 ~ Declaration: {normalArgDefault2.declaration()}")
    print(f"Normal Argument ~ Default ~ 1 ~ PyArg: {normalArgDefault1.pyarg()}")
    print(f"Normal Argument ~ Default ~ 2 ~ PyArg: {normalArgDefault2.pyarg()}")

    print(f"Const Argument ~ No Default ~ 1 ~ Param: {constArgNoDefault1.param()}")
    print(f"Const Argument ~ No Default ~ 2 ~ Param: {constArgNoDefault2.param()}")
    print(f"Const Argument ~ No Default ~ 1 ~ Declaration: {constArgNoDefault1.declaration()}")
    print(f"Const Argument ~ No Default ~ 2 ~ Declaration: {constArgNoDefault2.declaration()}")
    print(f"Const Argument ~ No Default ~ 1 ~ PyArg: {constArgNoDefault1.pyarg()}")
    print(f"Const Argument ~ No Default ~ 2 ~ PyArg: {constArgNoDefault2.pyarg()}")

    print(f"Ref Argument ~ No Default ~ 1 ~ Param: {refArgNoDefault1.param()}")
    print(f"Ref Argument ~ No Default ~ 2 ~ Param: {refArgNoDefault2.param()}")
    print(f"Ref Argument ~ No Default ~ 1 ~ Declaration: {refArgNoDefault1.declaration()}")
    print(f"Ref Argument ~ No Default ~ 2 ~ Declaration: {refArgNoDefault2.declaration()}")
    print(f"Ref Argument ~ No Default ~ 1 ~ PyArg: {refArgNoDefault1.pyarg()}")
    print(f"Ref Argument ~ No Default ~ 2 ~ PyArg: {refArgNoDefault2.pyarg()}")

    print(f"No Convert Argument ~ Default ~ 1 ~ Param: {noConvertArgDefault.param()}")
    print(f"No Convert Argument ~ Default ~ 1 ~ Declaration: {noConvertArgDefault.declaration()}")
    print(f"No Convert Argument ~ Default ~ 1 ~ PyArg: {noConvertArgDefault.pyarg()}")




# This file is part of the CLBlast project. The project is licensed under Apache Version 2.0. This file follows the
# PEP8 Python style guide and uses a max-width of 120 characters per line.
#
# Author(s):
#   Cedric Nugteren <www.cedricnugteren.nl>

NL = "\n"


def header():
    """Generates the header for the API documentation"""
    result = "CLBlast: API reference" + NL
    result += "================" + NL + NL + NL
    return result


def generate(routine):
    """Generates the API documentation for a given routine"""
    result = ""

    # Routine header
    result += "x" + routine.upper_name() + ": " + routine.description + NL
    result += "-------------" + NL + NL
    result += routine.details + NL + NL

    # Routine API
    result += "C++ API:" + NL
    result += "```" + NL
    result += routine.routine_header_cpp(12, "") + NL
    result += "```" + NL + NL
    result += "C API:" + NL
    result += "```" + NL
    for flavour in routine.flavours:
        result += routine.routine_header_c(flavour, 27, "") + NL
    result += "```" + NL + NL

    # Routine arguments
    result += "Arguments to " + routine.upper_name() + ":" + NL + NL
    for argument in routine.arguments_doc():
        result += "* " + argument + NL
    result += "* `cl_command_queue* queue`: "
    result += "Pointer to an OpenCL command queue associated with a context and device to execute the routine on." + NL
    result += "* `cl_event* event`: "
    result += "Pointer to an OpenCL event to be able to wait for completion of the routine's OpenCL kernel(s). "
    result += "This is an optional argument." + NL + NL

    # Routine requirements
    if len(routine.requirements_doc()) > 0:
        result += "Requirements for " + routine.upper_name() + ":" + NL + NL
        for requirement in routine.requirements_doc():
            result += "* " + requirement + NL
        result += NL

    # Routine footer
    result += NL + NL
    return result



import argument
import function
import class_
import module
import file

import itertools

# The set of Array types we support in Python
arrayTypes = []

for scalar in [("int16_t", "Int16"),
               ("int32_t", "Int32"),
               ("int64_t", "Int64"),
               ("float", "Float"),
               ("double", "Double"),
               ("lrc::Complex<float>", "ComplexFloat"),
               ("lrc::Complex<double>", "ComplexDouble")]:
    for backend in ["CPU", "OpenCL", "CUDA"]:
        arrayTypes.append({
            "scalar": scalar[0],
            "backend": backend,
            "name": f"Array{scalar[1]}{backend}"
        })


def generateCppArrayType(config):
    return f"lrc::Array<{config['scalar']}, lrc::backend::{config['backend']}>"


def generateCppArrayViewType(config):
    return f"lrc::array::GeneralArrayView<{generateCppArrayType(config)}>"


def generateFunctionsForArray(config):
    methods = [
        # Default constructor
        function.Function(
            name="__init__",
            args=[]
        )
    ]

    # Static fromData (n dimensions)
    for n in range(1, 9):
        cppType = ""
        for j in range(n):
            cppType += "std::vector<"
        cppType += config['scalar'] + ">" * n

        methods.append(
            function.Function(
                name="fromData",
                args=[
                    argument.Argument(
                        name=f"array{n}D",
                        type=cppType,
                        const=True,
                        ref=True,
                    )
                ],
                static=True,
                op=f"""
                    return {generateCppArrayType(config)}::fromData(array{n}D);
                """
            )
        )

    methods += [
        # Shape
        function.Function(
            name="__init__",
            args=[
                argument.Argument(
                    name="shape",
                    type="lrc::Shape",
                    const=True,
                    ref=True
                )
            ]
        ),

        # Shape and fill
        function.Function(
            name="__init__",
            args=[
                argument.Argument(
                    name="shape",
                    type="lrc::Shape",
                    const=True,
                    ref=True
                ),
                argument.Argument(
                    name="val",
                    type=config['scalar'],
                    const=True,
                    ref=True
                )
            ]
        ),

        # Copy constructor
        function.Function(
            name="__init__",
            args=[
                argument.Argument(
                    name="other",
                    type=generateCppArrayType(config),
                    const=True,
                    ref=True
                )
            ]
        ),

        # String representation
        function.Function(
            name="__str__",
            args=[
                argument.Argument(
                    name="self",
                    type=generateCppArrayType(config),
                    const=True,
                    ref=True
                )
            ],
            op="""
                return fmt::format("{}", self);
            """
        ),

        # String representation
        function.Function(
            name="__repr__",
            args=[
                argument.Argument(
                    name="self",
                    type=generateCppArrayType(config),
                    const=True,
                    ref=True
                )
            ],
            op=f"""
                return fmt::format("<librapid.{config['name']} ~ {{}}>", self.shape());
            """
        ),

        # Format (__format__)
        function.Function(
            name="__format__",
            args=[
                argument.Argument(
                    name="self",
                    type=generateCppArrayType(config),
                    const=True,
                    ref=True
                ),
                argument.Argument(
                    name="formatSpec",
                    type="std::string",
                    const=True,
                    ref=True
                )
            ],
            op="""
                std::string format = fmt::format("{{:{}}}", formatSpec);
                return fmt::format(fmt::runtime(format), self);
            """
        )
    ]

    return methods, []


def generateArrayModule(config):
    arrayClass = class_.Class(
        name=config["name"],
        type=generateCppArrayType(config)
    )

    methods, functions = generateFunctionsForArray(config)
    arrayClass.functions.extend(methods)

    includeGuard = None
    if config["backend"] == "CUDA":
        includeGuard = "defined(LIBRAPID_HAS_CUDA)"
    elif config["backend"] == "OpenCL":
        includeGuard = "defined(LIBRAPID_HAS_OPENCL)"

    arrayModule = module.Module(
        name=f"librapid.{config['name']}",
        includeGuard=includeGuard
    )
    arrayModule.addClass(arrayClass)
    arrayModule.functions.extend(functions)

    return arrayModule


def writeArray(root, config):
    fileType = file.File(
        path=f"{root}/{config['name']}.cpp"
    )

    fileType.modules.append(generateArrayModule(config))

    interfaceFunctions = fileType.write()
    # Run clang-format if possible
    try:
        import subprocess

        subprocess.run(["clang-format", "-i", fileType.path])
    except Exception as e:
        print("Unable to run clang-format:", e)

    return interfaceFunctions


def write(root):
    interfaces = []
    for config in arrayTypes:
        interfaces.extend(writeArray(root, config))
    return interfaces
